/* 
 * netra_watchdog.c - driver for hw watchdog inside Super I/O on NETRA AX-1105
 *
 * Copyright (C) 2002 Lineo (Embedix, Inc.)
 *
 * Driver was modified from David S. Miller's riowatchdog.c driver. 
 *
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/miscdevice.h>

#include <asm/io.h>
#include <asm/oplib.h>
#include <asm/uaccess.h>

#include <asm/mostek.h>
#include <asm/watchdog.h>

#define NVRAM_ADDR 	0x1fff1000000LL

#define WATCHDOG_STAT_ADDR	0x1fe010118d9LL
#define WATCHDOG_ADDR		0x1fe010118ccLL
#define WATCHDOG_TIMEOUT_BIT	0x20
#define MIN_TIME		3
#define MAX_TIME		63

MODULE_AUTHOR("Lineo, Embedix Inc. <www.lineo.com>");
MODULE_DESCRIPTION("Hardware watchdog driver for Sun NETRA AX-1105 ");
MODULE_SUPPORTED_DEVICE("netra_watchdog");

#define NETRAWD_NAME	"netra_watchdog"
#define NETRAWD_MINOR	215

/* Additional code for watchdog timer of Southbridge */

void set_wdt(u_char tval)
{
	tval |= 0x80; // for minutes timebase;
	mostek_write(WATCHDOG_ADDR,tval);
	mostek_write((WATCHDOG_ADDR + 3), 0x02); // enable the system reset and CPU reset
}

void set_watchdog_timeout(u_char tval)
{
	mostek_write(NVRAM_ADDR + 27, tval);
	set_wdt(tval);
}

u_char get_watchdog_timeout(void)
{
	u_char ret;

	ret = mostek_read(NVRAM_ADDR + 27);
	return (ret);
}

void set_watchdog_enable(u_char tval)
{
	mostek_write(NVRAM_ADDR + 14, tval);
}

u_char get_watchdog_enable(void)
{
	u_char ret;

	ret = mostek_read(NVRAM_ADDR + 14);
	return (ret);
}


static void wd_pingtimer(void)
{
	u_char tval;

	tval = get_watchdog_timeout();			
	set_watchdog_timeout(tval);
}

static void wd_stoptimer(void)
{
	set_watchdog_enable(0x00);
}

static void wd_starttimer(void)
{
	set_watchdog_enable(0xff);
}

static int wd_open(struct inode *inode, struct file *filp)
{
	return 0;
}

static int wd_release(struct inode *inode, struct file *filp)
{
	return 0;
}

static int wd_ioctl(struct inode *inode, struct file *filp,
		       unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case WDIOC_KEEPALIVE:
		wd_pingtimer();
		break;

	case WDIOC_SETOPTIONS:
		if (arg & WDIOS_DISABLECARD)
			wd_stoptimer();
		else if (arg & WDIOS_ENABLECARD)
			wd_starttimer();
		else
			return -EINVAL;

		break;

	default:
		return -EINVAL;
	};

	return 0;
}

static int wd_write(struct file *file, const char *buf, int count)
{
	set_watchdog_timeout((u_char)count);
	return 0;
}

static ssize_t wd_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
	return -EINVAL;
}

static struct file_operations wd_fops = {
	owner:		THIS_MODULE,
	ioctl:		wd_ioctl,
	open:		wd_open,
	write:		wd_write,
	read:		wd_read,
	release:	wd_release,
};

static struct miscdevice wd_miscdev = { NETRAWD_MINOR, NETRAWD_NAME, &wd_fops };

static int __init wd_init(void)
{
	u_char status, time_in_min;

	status = mostek_read(WATCHDOG_STAT_ADDR);
	if (status & WATCHDOG_TIMEOUT_BIT)
	{
		printk("\n WDT: System reset due to Watchdog timeout \n");
		mostek_write(WATCHDOG_STAT_ADDR, WATCHDOG_TIMEOUT_BIT);
	}


	if (get_watchdog_enable())
	{
		time_in_min = get_watchdog_timeout();
		if (time_in_min < MIN_TIME)
		{
			printk("\nWDT setting: Setting WD timeout to %d",MIN_TIME);
			time_in_min = MIN_TIME;
			set_watchdog_timeout(MIN_TIME);
		}
		if (time_in_min > MAX_TIME)
		{
			printk("\nWDT setting: Setting WD timeout to %d",MAX_TIME);
			time_in_min = MAX_TIME;
			set_watchdog_timeout(MAX_TIME);
		}
		
	}
	else
		set_watchdog_enable(0x00); // Disable the watchdog 

	if (misc_register(&wd_miscdev)) {
		printk(KERN_ERR "watchdog: Cannot register watchdog misc device.\n");
		return -1;
	}
	return 0;
}

static void __exit wd_cleanup(void)
{
	misc_deregister(&wd_miscdev);
}

module_init(wd_init);
module_exit(wd_cleanup);



