/* $Id: rtc.c,v 1.26 2001/03/14 09:30:31 davem Exp $
 *
 * Linux/SPARC Real Time Clock Driver
 * Copyright (C) 1996 Thomas K. Dyas (tdyas@eden.rutgers.edu)
 *
 * This is a little driver that lets a user-level program access
 * the SPARC Mostek real time clock chip. It is no use unless you
 * use the modified clock utility.
 *
 * Get the modified clock utility from:
 *   ftp://vger.kernel.org/pub/linux/Sparc/userland/clock.c
 */

/* This driver was modified for sparc64 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <asm/mostek.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/rtc.h>

static int rtc_busy = 0;

#define TRUE 1
#define FALSE 0
#define NVRAM_ADDR	0x1fff1000000LL
#define NVRAM_SIZE	0x2000
#define TIMER_ADDR	(NVRAM_ADDR + NVRAM_SIZE - 8)

/* Retrieve the current date and time from the real time clock. */
void get_rtc_time(struct rtc_time *t)
{
	unsigned long regs = mstk48t02_regs;
	u8 tmp;

	get_time(&t->hour,&t->min,&t->sec);
	get_date(&t->month,&t->dom,&t->year,&t->dow);
}

/* Set the current date and time inthe real time clock. */
void set_rtc_time(struct rtc_time *t)
{
	unsigned long regs = mstk48t02_regs;
	u8 tmp;

	set_time(t->hour,t->min,t->sec);
	set_date(t->month,t->dom,t->year,t->dow);
}

static long long rtc_lseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}

static int rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{
	struct rtc_time rtc_tm;

	switch (cmd)
	{
	case RTCGET:
		get_rtc_time(&rtc_tm);

		if (copy_to_user((struct rtc_time*)arg, &rtc_tm, sizeof(struct rtc_time)))
			return -EFAULT;

		return 0;


	case RTCSET:
		/*if (!capable(CAP_SYS_TIME))
			return -EPERM;*/

		if (copy_from_user(&rtc_tm, (struct rtc_time*)arg, sizeof(struct rtc_time)))
			return -EFAULT;

		set_rtc_time(&rtc_tm);

		return 0;

	default:
		return -EINVAL;
	}
}

static int rtc_open(struct inode *inode, struct file *file)
{
	int ret;

	spin_lock_irq(&mostek_lock);
	if (rtc_busy) {
		ret = -EBUSY;
	} else {
		rtc_busy = 1;
		ret = 0;
	}
	spin_unlock_irq(&mostek_lock);

	return ret;
}

static int rtc_release(struct inode *inode, struct file *file)
{
	rtc_busy = 0;

	return 0;
}

static struct file_operations rtc_fops = {
	owner:		THIS_MODULE,
	llseek:		rtc_lseek,
	ioctl:		rtc_ioctl,
	open:		rtc_open,
	release:	rtc_release,
};

static struct miscdevice rtc_dev = { RTC_MINOR, "rtc", &rtc_fops };

EXPORT_NO_SYMBOLS;

static int __init rtc_sun_init(void)
{
	int error;

	/* It is possible we are being driven by some other RTC chip
	 * and thus another RTC driver is handling things.
	 */
	/*if (mstk48t02_regs == 0)
		return -ENODEV;*/

	new_timer();

	error = misc_register(&rtc_dev);
	if (error) {
		printk(KERN_ERR "rtc: unable to get misc minor for Mostek\n");
		return error;
	}

	return 0;
}

static void __exit rtc_sun_cleanup(void)
{
	misc_deregister(&rtc_dev);
}

module_init(rtc_sun_init);
module_exit(rtc_sun_cleanup);


 /* Additional code for UltarSparcIIe - TOD 48T59 chip */

void
timer_store (u_char data, int offset)
{
	mostek_write((TIMER_ADDR+offset),data);
}

u_char
timer_load (int offset)
{
	return (mostek_read((TIMER_ADDR+offset)));
}

void
bcd_timer_store (u_char data, int offset)
{
	u_char quot, rem;
	quot = data / 10;
	rem = data % 10;
	data = (quot << 4) + rem;
	timer_store (data, offset);
}

u_char
bcd_timer_load (int offset)
{
	u_char low, hi;
	u_char data = timer_load (offset);

	low = data & 0x0f;
	hi  = ((data & 0xf0) >> 4);
	return ((hi * 10) + low);
}

void
stop_rtc (void)
{
	timer_store (0x80, 0);		// Write bit on
	timer_store (0x80, 1);		// Stop bit on
	timer_store (0x00, 0);		// Write bit off
}

void
start_rtc (void)
{
	int i;

	timer_store (0x80, 0);		// Write bit on
	timer_store (0x00, 1);		// Stop bit off
	timer_store (0x00, 0);		// Write bit off
	for (i=0; i<100000; i++);	// delay
	timer_store (0x80, 0);		// Write bit on again
	timer_store (0x80, 0);		// Write bit on again
	bcd_timer_store (0, 7);		// Year
	bcd_timer_store (1, 6);		// Month
	bcd_timer_store (1, 5);		// Date
	bcd_timer_store (1, 4);		// Day of week
	bcd_timer_store (0, 3);		// Hour
	bcd_timer_store (0, 2);		// Minute
	bcd_timer_store (0, 1);		// Seconds
	timer_store (0x00, 0);		// Write bit off
}

int
get_seconds (void)
{
	u_char sec;

	timer_store (0x00, 0);		// Turn off read bit to "unfreeze" the counter
	timer_store (0x40, 0);		// "Read bit" freezes values
	sec = bcd_timer_load (1);	// seconds
	timer_store (0x00, 0);		// Unfreeze values
	return ((int) sec);
}

void
get_date (int *month, int *date, int *year, int *dow)
{
	timer_store (0x00, 0);		// Turn off read bit to "unfreeze" the counter
	timer_store (0x40, 0);		// "Read bit" freezes values
	*month = (int) bcd_timer_load (6);	// month
	*date  = (int) bcd_timer_load (5);	// date
	*year  = (int) bcd_timer_load (7);	// year
	*year += 1968;
	*dow = (int) (0x07 & bcd_timer_load(4));
	timer_store (0x00, 0);		// Unfreeze values
}

void
get_time (int *hour, int *min, int *secs)
{
	timer_store (0x00, 0);		// Turn off read bit to "unfreeze" the counter
	timer_store (0x40, 0);		// "Read bit" freezes values
	*hour = (int) bcd_timer_load (3);	// hour
	*min  = (int) bcd_timer_load (2);	// minutes
	*secs = (int) bcd_timer_load (1);	// seconds
	timer_store (0x00, 0);		// Unfreeze values
}

int
set_date (int month, int date, int year, int dow)
{
	int i;

	year -= 1968;
	if ((month < 1) || (month > 12))
		return (TRUE);
	if ((date < 1) || (date > 31))
		return (TRUE);
	if (year < 0)
		return (TRUE);

	timer_store (0x80, 0);		// Write bit on again
	bcd_timer_store ((u_char) month, 6);	// month
	bcd_timer_store ((u_char) date, 5);	// date
	bcd_timer_store ((u_char) year, 7);	// year
	bcd_timer_store((u_char) dow, 4); 	//dow
	timer_store (0x00, 0);		// Write bit off
}

int
set_time (int hour, int min, int secs)
{
	int i;

	if ((hour < 0) || (hour > 23))
		return (TRUE);
	if ((min < 0) || (min > 59))
		return (TRUE);
	if ((secs < 0) || (secs > 59))
		return (TRUE);

	timer_store (0x80, 0);		// Write bit on again
	bcd_timer_store ((u_char) hour, 3);		// hour
	bcd_timer_store ((u_char) min, 2);		// minutes
	bcd_timer_store ((u_char) secs, 1);		// seconds
	timer_store (0x00, 0);		// Write bit off

	return (FALSE);
}

int
clock_ticking (void)
{
	int old, new, i, j;

	for (i=0; i<30; i++)
	{
		old = get_seconds ();		// get seconds reading
		for (j=0; j<10000; j++);	// delay
		new = get_seconds ();		// get seconds reading again
		if (old != new)				// Clock changed!
			return (TRUE);			//   Its ticking
		for (j=0; j<10000; j++);	// delay for another reading
	}
	return (FALSE);					// Not ticking
}

void
new_timer (void)
{
	if (!clock_ticking)
		start_rtc ();
}

